#include "../header/hashing.h"
#include "../header/config.h"
#include "../header/fnv.h"
#include <stdio.h>

uint32 roll_hashx(unsigned char c, uchar window[], uint32 rhData[]) {
	uint32 t = rhData[0] % ROLLING_WINDOW;

	rhData[2] -= rhData[1];
	rhData[2] += (ROLLING_WINDOW * c);

	rhData[1] += c;
	rhData[1] -= window[t];

	window[t] = c;
	rhData[0]++;

	rhData[3] = (rhData[3] << 5); //& 0xFFFFFFFF;
	rhData[3] ^= c;

	return rhData[1] + rhData[2] + rhData[3];
}

//Depending on the third parameter it fills the Bloom filter (1) or it compares File against Bloom filter
//doWhat: 1 = hashAndAdd  ; 2 = hashAndCompare ; 3 = remove
int *hashFileAndDo(BLOOMFILTER *bf, unsigned char *byte_buffer,
		unsigned char* cbf, int doWhat, unsigned int start, unsigned int stop) {
	unsigned int i;
	uint32 rValue;

	unsigned int randomness = 0;

	unsigned int last_block_index = start;
	float entropy_val = 0.0, tmp_entro = 0.0;

	unsigned int *frequencyArray = (unsigned int *) calloc(256, sizeof(int));
	//memset(frequencyArray,0,sizeof(int)*256);
	unsigned int *results_summary = (unsigned int *) calloc(4, sizeof(int)); //0 total blocks; 1 blocks found; 2 longest run; 3 tmp savings
	//memset(results_summary,0, sizeof(int)*4 );

	unsigned int *hash_val;
	unsigned int skip_counter = 0;

	/*we need this arrays for our extended rollhash function*/
	uchar window[ROLLING_WINDOW] = { 0 };
	uint32 rhData[4] = { 0 };

	//run through all bytes
	for (i = start; i < stop; i++) {
		//update frequency array for entropy
		frequencyArray[byte_buffer[i]]++;

		//Randomness
		if(i+1 < stop && abs(byte_buffer[i]-byte_buffer[i+1]) <= 1)
			randomness++;

		//skip rolling hash if chunk is too small
		if(skip_counter++ < SKIPPED_BYTES)
			continue;

		//build rolling hash
		rValue = roll_hashx(byte_buffer[i], window, rhData);

		//check for end of chunk
		if (rValue % BLOCK_SIZE == BLOCK_SIZE - 1) {
			//a chunk was found, so increase counter
			results_summary[0]++;

			//calculate entropy of piece
			entropy_val = (double) entropy(frequencyArray,
					i - last_block_index);

			if (entropy_val == tmp_entro)
				continue;


			//if entropy is over our MIN_ENTROPY, hash and add ....
			if ((entropy_val >= MIN_ENTROPY) && (randomness*3 < (i - last_block_index))) {
				//hash the chunk and add hash to bloom filter
				hash_val = fnv256Bit(byte_buffer, last_block_index, i);


				//check whether we are in comparison mode or checking mode
				if (doWhat == 1){
					add_hash_to_bloomfilter(bf, hash_val, cbf);
					//if(hash_val[0]==992172228)
						//printf("bo\n");
				}else if (doWhat == 2) {
					//createResultsSummary(bf, hash_val, results_summary);	//check whether we are in comparison mode or checking mode
					if (createResultsSummary(bf, hash_val, results_summary)
							&& results_summary[2] > 5) {
							//for(int x=0; x<60; x++)
								//printf("%x",byte_buffer[x+last_block_index]);
							//printf("%u ", hash_val[0]);
						//printf("%i < %i (e: %f) ",randomness*2, i - last_block_index, entropy_val);
						//hexDump ("byte_buffer", &byte_buffer[last_block_index], i-last_block_index);
					}
				} else
					remove_hash_from_filter(bf, hash_val, cbf);
				free(hash_val);

			} else {	//...otherwise reset counter for longest run
				results_summary[3] = 0;
			}

			//set stuff for next round
			tmp_entro = entropy_val;
			last_block_index = i + 1;

			//reset counter
			skip_counter = 0;
			randomness = 0;
		}
	}

	free(frequencyArray);
	return results_summary;
}

double entropy(int freqArray[], int size) {
	double e = 0.0;
	double f = 0.0;
	int i = 0;
	for (i = 0; i <= 255; i++) {
		if (freqArray[i] > 0) {
			f = (double) freqArray[i] / size;
			e -= f * log2(f);
			freqArray[i] = 0; 		//reset
		}
	}
	return e;
}

int createResultsSummary(BLOOMFILTER *bf, uint256 hash_val,
		int *results_summary) {
	if (is_in_bloom(bf, hash_val) == 1) {
		results_summary[1]++;					//counter for found chunks
		results_summary[3]++;
		if (results_summary[3] > results_summary[2]) //check if there is a long run
			results_summary[2] = results_summary[3];
		return 1;
	} else
		results_summary[3] = 0;
	return 0;
}

void hexDump(char *desc, void *addr, int len) {
	int i;
	unsigned char buff[17];
	unsigned char *pc = addr;

	// Output description if given.
	if (desc != NULL)
		printf("%s:\n", desc);

	// Process every byte in the data.
	for (i = 0; i < len; i++) {
		// Multiple of 16 means new line (with line offset).

		if ((i % 16) == 0) {
			// Just don't print ASCII for the zeroth line.
			if (i != 0)
				printf("  %s\n", buff);

			// Output the offset.
			printf("  %04x ", i);
		}

		// Now the hex code for the specific character.
		printf(" %02x", pc[i]);

		// And store a printable ASCII character for later.
		if ((pc[i] < 0x20) || (pc[i] > 0x7e))
			buff[i % 16] = '.';
		else
			buff[i % 16] = pc[i];
		buff[(i % 16) + 1] = '\0';
	}

	// Pad out last line if not exactly 16 characters.
	while ((i % 16) != 0) {
		printf("   ");
		i++;
	}

	// And print the final ASCII bit.
	printf("  %s\n", buff);
}
