/* 
 * File:   main.c
 * Author: Frank Breitinger
 * Created on 28. April 2013, 19:15
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <dirent.h>
#include <openssl/md5.h>
#include <limits.h>		/* PATH_MAX */
#include "../header/main.h"
#include "../header/helper.h";

// Global variable for the different modes
MODES *mode;

extern char *optarg;
extern int optind;
int USE_COUNTING_BF = 1;
float MIN_ENTROPY = 0.0;
char *filetype;
static int counter=0;


static void show_help(void) {
	printf(
			"\nmrsh-v2  by Frank Breitinger\n"
					"Copyright (C) 2013 \n"
					"\n"
					"Usage: mrsh-net [-gh] [-s size] DIR/FILE \n"
					"OPTIONS: -g: Generate the database and print it to std (save to a file like: './mrsh-net DIR > myDB' \n"
					"             if unsetbits is used it uset them\n"
					"         -f: generates the false positive for a list of files; call: ./mrsh-net -f mydir/* \n"
					"         -i: Reads DB-FILE and compares DIR/FILE against it. \n"
					"         -e: sets the minimum entropy\n"
					"         -t  excludes a filetype, e.g., './mrsh-net -t .jpg -g DIR' filetype has to start with a '.' \n"
					"         -r: Reads directory recursive, default no \n"
					"         -h: Print this help message \n\n");
}

static void initalizeDefaultModes() {
	mode = (MODES *) malloc(sizeof(MODES));
	mode->helpmessage = false;
	mode->generateBF = false;
	mode->recursive = true;
	mode->fp_mode = false;
	mode->readDB = false;
	mode->all_a_all = false;

}

int main(int argc, char **argv) {

	int i;
	unsigned char *cbf;
	unsigned long cbf_size = (unsigned int) BF_SIZE_IN_BYTES * 8; //for all bits we need 1 byte ... *8
	initalizeDefaultModes();
	initialize_settings();		//Bloom filter settings
	filetype = malloc(5);
	filetype = ".foo";

	char *listName = NULL;

	while ((i = getopt(argc, argv, "i:ghrfe:t:a:")) != -1) {
		switch (i) {
		case 'a':
			listName = optarg;
			mode->all_a_all = true;
			break;
		case 'i':
			mode->readDB = true;
			listName = optarg;
			break;
		case 'g':
			mode->generateBF = true;
			break;
		case 'f':
			mode->fp_mode = true;
			break;
		case 'e':
			MIN_ENTROPY = atof(optarg);
			break;
		case 't':
			filetype = optarg;
			break;
		case 'r':
			mode->recursive = true;
			break;
		case 'h':
			mode->helpmessage = true;
			break;
		default:
			mode->helpmessage = true;
			fprintf(stderr, "[*] Unknown option(s) \n");
			break;
		}
	}

	if (mode->helpmessage) {
		show_help();
		exit(0);
	}

	//read db from file and compare stuff against it
	if (mode->readDB) {
		USE_COUNTING_BF = 0;
		BLOOMFILTER *bf = init_empty_BF();

		readFileToBF(listName, bf);
		unsigned char *buffer;
		int size;
		//printf("i-mode (DB: %s) - MIN_ENTROPY: %.2f, UNSET_BITS_THRES: %i, MIN_RUN: %i, BLOCK_SIZE: %i \n", listName, MIN_ENTROPY, UNSET_BITS_THRES, MIN_RUN, BLOCK_SIZE);

		//read in the "DB"
		for (int j = optind; j < argc; j++) {
			FILE *file = getFileHandle(argv[j]);
			size = find_file_size(file);
			buffer = getByteBuffer(file);
			evaluation(bf, argv[j], buffer, size);
			free(buffer);
			fclose(file);
		}

		destroy_bf(bf);
		exit(1);
	}

	//read all arguments, create the bloomfilter, and print it to stdout
	if (mode->generateBF || optind == 1) {
		//Create new BF including setting counting Bloom filter
		BLOOMFILTER *bf = init_empty_BF();

		if (USE_COUNTING_BF == 1) {
			cbf = (unsigned char*) malloc(cbf_size);
			memset(cbf, 0, cbf_size);
		}

		int size;
		unsigned char *buffer;
		//fill bloom filter with all files
		for (int j = optind; j < argc; j++) {
			if (endsWithType(argv[j])) {
				continue;
			}
			//printf("%s \n",argv[j]);
			FILE *file = getFileHandle(argv[j]);
			size = find_file_size(file);
			buffer = getByteBuffer(file);

			hashFileAndDo(bf, buffer, cbf, 1, 0, size);

			free(buffer);
			fclose(file);
		}

		//unsetting all bits that set to often (over threshold)
/*		unsigned distri[256] = {0};
		unsigned long overall = 0;
		for (int i = 1; i < cbf_size; i++) {
			distri[cbf[i]]++;
			overall += cbf[i];
		}
		double amount=0.0;
		for (int i = 0; i < 256; i++) {
			amount +=((double)(distri[i]))/(double)overall;
			if(distri[i] > 0)
				printf("%i: %i (%f)\n", i, distri[i], amount);
		}

*/
		//unsetting all bits that set to often (over threshold)
		for (int i = 0; i < cbf_size; i++) {
			if (cbf[i] >= UNSET_BITS_THRES)
				unset_bit(bf, i);
		}


		free(cbf);
		print_bf(bf);
		destroy_bf(bf);
		exit(1);
	}



	//read all arguments, create the bloomfilter, and print it to stdout
	if (mode->fp_mode || optind == 1) {

		BLOOMFILTER *bf = init_empty_BF();
		if (USE_COUNTING_BF == 1)
			cbf = (unsigned char*) calloc(1, cbf_size);


		int size;
		unsigned char *buffer;
		//fill bloom filter with all files
		for (int j = optind; j < argc; j++) {
			if (endsWithType(argv[j]))
				continue;
			FILE *file = getFileHandle(argv[j]);
			size = find_file_size(file);
			buffer = getByteBuffer(file);

			hashFileAndDo(bf, buffer, cbf, 1, 0, size);

			free(buffer);
			fclose(file);
		}

		//unsetting all bits that set to often (over threshold)
	/*	unsigned distri[256] = {0};
		unsigned long overall = 0;
		for (int i = 1; i < cbf_size; i++) {
			distri[cbf[i]]++;
			overall += cbf[i];
		}
		float amount=0.0;
		for (int i = 1; i < 256; i++) {
			amount +=((float)(distri[i]))/(float)overall;
			printf("%i: %i (%f)\n", i, distri[i], amount);
		}
*/


		//unsetting all bits that set to often (over threshold)
		for (int i = 0; i < cbf_size; i++) {
			if (cbf[i] >= UNSET_BITS_THRES)
				unset_bit(bf, i);
		}
		//do false positive test by using counting bloom filter to delete a file and compare it afterwards
		//printf("fp-mode - MIN_ENTROPY: %.2f, UNSET_BITS_THRES: %i, MIN_RUN: %i, BLOCK_SIZE: %i (filetyp: %s) \n", MIN_ENTROPY, UNSET_BITS_THRES, MIN_RUN, BLOCK_SIZE, filetype);
		for (int j = optind; j < argc; j++) {
			if (!endsWithType(argv[j]) && !endsWithType(".foo"))
				continue;

			FILE *file = getFileHandle(argv[j]);
			size = find_file_size(file);
			buffer = getByteBuffer(file);

			hashFileAndDo(bf, buffer, cbf, 3, 0, size);
			evaluation(bf, argv[j], buffer, size);
			hashFileAndDo(bf, buffer, cbf, 1, 0, size);

			free(buffer);
			fclose(file);
		}

		destroy_bf(bf);
		exit(1);
	}


	//read all arguments, create the bloomfilter, and print it to stdout
	if (mode->all_a_all) {
		BLOOMFILTER *bf = init_empty_BF();
		readFileToBF(listName, bf);

		if (USE_COUNTING_BF == 1)
					cbf = (unsigned char*) calloc(1, cbf_size);


		unsigned char *buffer;
		int size;

		BLOOMFILTER *tmp_BF = init_empty_BF();

		//fill bloom filter with all files
		for (int j = optind; j < argc; j++) {
			FILE *file = getFileHandle(argv[j]);
			size = find_file_size(file);
			buffer = getByteBuffer(file);

			hashFileAndDo(tmp_BF, buffer, cbf, 1, 0, size);

			free(buffer);
			fclose(file);
		}
		destroy_bf(tmp_BF);


		//unsetting all bits that set to often (over threshold)
		for (int i = 0; i < cbf_size; i++) {
			if (cbf[i] >= UNSET_BITS_THRES)
				unset_bit(bf, i);
		}

		//compare files to DB
		for (int j = optind; j < argc; j++) {
			FILE *file = getFileHandle(argv[j]);
			size = find_file_size(file);
			buffer = getByteBuffer(file);
			evaluation(bf, argv[j], buffer, size);
			free(buffer);
			fclose(file);
		}

		destroy_bf(bf);
		exit(1);
	}





	//combine x filters
/*	if (mode->combine2filter) {
		//Create new BF including setting counting Bloom filter
		BLOOMFILTER *bf = init_empty_BF();
		BLOOMFILTER *bfe = init_empty_BF();

		for (int j = optind; j < argc; j++) {
			readFileToBF(argv[j], bfe);
			for (i = 0; i < BF_SIZE_IN_BYTES; i++) {
				bf->array[i] = bf->array[i] | bfe->array[i];
			}
		}

		print_bf(bf);
		destroy_bf(bf);
		exit(1);
	}
*/


}

/*
 * adds a path to a fingerprints list. may be recursive depending on the parameters
 */
void addPathToBloomfilter(BLOOMFILTER *bf, char *filename, unsigned char* cbf) {
	DIR *dir;
	struct dirent *ent;
	const int max_path_length = 1024;
	unsigned char *buffer;
	int size;

	char *cur_dir = (char *) malloc(max_path_length);
	getcwd(cur_dir, max_path_length);

	//in case of a dir
	if (is_dir(filename)) {
		dir = opendir(filename);
		chdir(filename);

		//run through all files of the dir
		while ((ent = readdir(dir)) != NULL) {

			//if we found a file, generate hash value and add it
			if (is_file(ent->d_name)) {
				FILE *file = getFileHandle(ent->d_name);
				size = find_file_size(file);
				buffer = getByteBuffer(file);
				hashFileAndDo(bf, buffer, cbf, 1, 0, size);
				free(buffer);
				fclose(file);
				//FINGERPRINT *fp = init_fingerprint_for_file(file, ent->d_name);
				//add_new_fingerprint(bf, fp);
			}

			//when we found a dir and recursive mode is on, go deeper
			else if (is_dir(ent->d_name) && mode->recursive) {
				if (strcmp(ent->d_name, ".") == 0
						|| strcmp(ent->d_name, "..") == 0)
					continue;
				addPathToBloomfilter(bf, ent->d_name, cbf);
			}
		}
		chdir(cur_dir);
		closedir(dir);
	}

	//in case we we have only a file
	else if (is_file(filename)) {
		FILE *file = getFileHandle(filename);
		size = find_file_size(file);
		buffer = getByteBuffer(file);
		hashFileAndDo(bf, buffer, cbf, 1, 0, size);
		free(buffer);
		fclose(file);
	}

	return;
}

void evaluation(BLOOMFILTER *bf, char *filename, unsigned char *buffer,
		int size) {
	int *results;
	int ss = 0, c = 0;
	int flag = 0;


	for (ss = 0, c = 0; ss + PACKET_SIZE < size; ss = ss + PACKET_SIZE, c++) {
		results = hashFileAndDo(bf, buffer, NULL, 2, ss, ss + PACKET_SIZE - 1);
		if (results[2] >= MIN_RUN) {
			//printf("%s Packet %i: %i of %i (longest run: %i)\n", filename, c, results[1], results[0], results[2]);
			flag++;
		}
		counter++;

		free(results);
	}
	c++;
	results = hashFileAndDo(bf, buffer, NULL, 2, ss, size);
	if (results[2] >= MIN_RUN) {
		//printf("%s Packet %i: %i of %i (longest run: %i)\n", filename, c, results[1], results[0], results[2]);
		flag++;
	}
	counter++;

	free(results);

	if (flag >= 1) {
		printf("%s %i of %i  \n", filename, flag, c);
	}

}

void evaluation2(BLOOMFILTER *bf, char *filename, unsigned char *buffer,
		int size) {
	unsigned int *results;
	int ss = 0, c = 0;

	for (ss = 0, c = 0; ss + PACKET_SIZE < size; ss = ss + PACKET_SIZE, c++) {
		results = hashFileAndDo(bf, buffer, NULL, 2, ss, ss + PACKET_SIZE - 1);
		if (results[2] >= MIN_RUN) {

		printf("%s, %i, %i, %i, %i\n", filename, c, results[1], results[0],
				results[2]);
		}
		free(results);
	}
	c++;
	results = hashFileAndDo(bf, buffer, NULL, 2, ss, size);
	printf("%s, %i, %i, %i, %i\n", filename, c, results[1], results[0],
			results[2]);
	free(results);
}

int endsWithType(char *str) {
	char *dot = strrchr(str, '.');
	return strcmp(dot, filetype) == 0;
}

